import librosa
import soundfile as sf
import random
import numpy as np
import jax
import jax.numpy as jnp
from multiprocessing import pool
from typing import Optional
import munch

from flaxspeaker import dataset
from flaxspeaker import specaug


def extract_features(audio_file: str, n_mfcc: int) -> np.ndarray:
    """Extract MFCC features from an audio file, shape=(TIME, MFCC)."""
    waveform, sample_rate = sf.read(audio_file)

    # Convert to mono-channel.
    if len(waveform.shape) == 2:
        waveform = librosa.to_mono(waveform.transpose())

    # Convert to 16kHz.
    if sample_rate != 16000:
        waveform = librosa.resample(waveform, sample_rate, 16000)

    features = librosa.feature.mfcc(
        y=waveform, sr=sample_rate, n_mfcc=n_mfcc)
    return features.transpose()


def extract_sliding_windows(features: np.ndarray,
                            myconfig: munch.Munch) -> list[np.ndarray]:
    """Extract sliding windows from features."""
    sliding_windows = []
    start = 0
    while start + myconfig.model.seq_len <= features.shape[0]:
        sliding_windows.append(
            features[start: start + myconfig.model.seq_len, :])
        start += myconfig.model.sliding_window_step
    return sliding_windows


def get_triplet_features(spk_to_utts: dataset.SpkToUtts,
                         n_mfcc: int
                         ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Get a triplet of anchor/pos/neg features."""
    anchor_utt, pos_utt, neg_utt = dataset.get_triplet(spk_to_utts)
    return (extract_features(anchor_utt, n_mfcc),
            extract_features(pos_utt, n_mfcc),
            extract_features(neg_utt, n_mfcc))


def trim_features(features: np.ndarray,
                  seq_len: int,
                  specaug_config: munch.Munch
                  ) -> np.ndarray:
    """Trim features to SEQ_LEN."""
    full_length = features.shape[0]
    start = random.randint(0, full_length - seq_len)
    trimmed_features = features[start: start + seq_len, :]
    if specaug_config.use_specaug:
        trimmed_features = specaug.apply_specaug(
            trimmed_features, specaug_config)
    return trimmed_features


class TrimmedTripletFeaturesFetcher:
    """The fetcher of trimmed features for multi-processing."""

    def __init__(self,
                 spk_to_utts: dataset.SpkToUtts,
                 myconfig: munch.Munch):
        self.spk_to_utts = spk_to_utts
        self.seq_len = myconfig.model.seq_len
        self.specaug_config = myconfig.train.specaug
        self.n_mfcc = myconfig.model.n_mfcc

    def __call__(self, _) -> np.ndarray:
        """Get a triplet of trimmed anchor/pos/neg features."""
        anchor, pos, neg = get_triplet_features(self.spk_to_utts, self.n_mfcc)
        while (anchor.shape[0] < self.seq_len or
               pos.shape[0] < self.seq_len or
               neg.shape[0] < self.seq_len):
            anchor, pos, neg = get_triplet_features(
                self.spk_to_utts, self.n_mfcc)
        return np.stack([
            trim_features(anchor, self.seq_len, self.specaug_config),
            trim_features(pos, self.seq_len, self.specaug_config),
            trim_features(neg, self.seq_len, self.specaug_config)])


def get_batched_triplet_input(
        spk_to_utts: dataset.SpkToUtts,
        myconfig: munch.Munch,
        pool: Optional[pool.Pool] = None) -> jax.Array:
    """Get batched triplet input for Jax."""
    fetcher = TrimmedTripletFeaturesFetcher(spk_to_utts, myconfig)
    if pool is None:
        input_arrays = list(map(fetcher, range(myconfig.train.batch_size)))
    else:
        input_arrays = pool.map(fetcher, range(myconfig.train.batch_size))
    batch_input = np.concatenate(input_arrays)
    return jnp.asarray(batch_input)



import numpy as np
import random
import munch


def apply_specaug(
        features: np.ndarray,
        specaug_config: munch.Munch) -> np.ndarray:
    """Apply SpecAugment to features."""
    seq_len, n_mfcc = features.shape
    outputs = features
    mean_feature = np.mean(features)

    # Frequancy masking.
    if random.random() < specaug_config.freq_mask_prob:
        width = random.randint(1, specaug_config.freq_mask_max_width)
        start = random.randint(0, n_mfcc - width)
        outputs[:, start: start + width] = mean_feature

    # Time masking.
    if random.random() < specaug_config.time_mask_prob:
        width = random.randint(1, specaug_config.time_mask_max_width)
        start = random.randint(0, seq_len - width)
        outputs[start: start + width, :] = mean_feature

    return outputs



import os
import time
import jax
import jax.numpy as jnp
import flax
from flax.training import train_state
from flax import linen as nn
import tensorflow as tf
import optax
import matplotlib.pyplot as plt
import multiprocessing
from typing import Optional
import munch
from functools import partial
import sys

from flaxspeaker import dataset
from flaxspeaker import feature_extraction


class BaseSpeakerEncoder(nn.Module):
    pass


class LstmSpeakerEncoder(BaseSpeakerEncoder):
    lstm_config: munch.Munch

    def setup(self):
        self.lstm_layers = [
            nn.RNN(nn.OptimizedLSTMCell(
                features=self.lstm_config["hidden_size"]))
            for i in range(self.lstm_config["num_layers"])]

    def _aggregate_frames(self, batch_output: jax.Array) -> jax.Array:
        """Aggregate output frames."""
        if self.lstm_config["frame_aggregation_mean"]:
            return jnp.mean(
                batch_output, axis=1, keepdims=False)
        else:
            return batch_output[:, -1, :]

    def __call__(self, x: jax.Array) -> jax.Array:
        for lstm in self.lstm_layers:
            x = lstm(x)
        return self._aggregate_frames(x)


class TransformerSpeakerEncoder(BaseSpeakerEncoder):
    transformer_config: munch.Munch

    def setup(self):
        # Define the Transformer network.
        self.linear_layer = nn.Dense(features=self.transformer_config["dim"])
        self.encoders = [
            nn.SelfAttention(num_heads=self.transformer_config["num_heads"])
            for i in range(self.transformer_config["num_encoder_layers"])]
        self.temporal_attention = nn.Dense(features=1)

    def __call__(self, x: jax.Array) -> jax.Array:
        encoder_input = nn.activation.sigmoid(self.linear_layer(x))
        for encoder in self.encoders:
            encoder_output = encoder(encoder_input)
            encoder_input = encoder_output

        # Attentive temporal pooling.
        temporal_weights = self.temporal_attention(encoder_output)
        weighted_output = jnp.multiply(encoder_output, temporal_weights)
        weighted_output = jax.nn.softmax(weighted_output)
        return jnp.mean(weighted_output, axis=1, keepdims=False)


@jax.jit
def cosine_similarity(a: jax.Array, b: jax.Array) -> jax.Array:
    """Compute cosine similarity between two embeddings."""
    eps = 1e-6
    return jnp.dot(a, b) / (jnp.linalg.norm(a) * jnp.linalg.norm(b) + eps)


@jax.jit
def get_triplet_loss(anchor: jax.Array, pos: jax.Array, neg: jax.Array,
                     alpha: float) -> jax.Array:
    """Triplet loss defined in https://arxiv.org/pdf/1705.02304.pdf."""

    return jnp.maximum(
        jax.vmap(cosine_similarity, in_axes=[0, 0])(anchor, neg) -
        jax.vmap(cosine_similarity, in_axes=[0, 0])(anchor, pos) +
        alpha,
        0.0)


@partial(jax.jit, static_argnames=["batch_size", "triplet_alpha"])
def get_triplet_loss_from_batch_output(batch_output: jax.Array,
                                       batch_size: int,
                                       triplet_alpha: float
                                       ) -> jax.Array:
    """Triplet loss from N*(a|p|n) batch output."""
    batch_output_reshaped = jnp.reshape(
        batch_output, (batch_size, 3, batch_output.shape[1]))
    batch_loss = get_triplet_loss(
        batch_output_reshaped[:, 0, :],
        batch_output_reshaped[:, 1, :],
        batch_output_reshaped[:, 2, :],
        triplet_alpha)
    loss = jnp.mean(batch_loss)
    return loss


def save_model(saved_model_path: str, state: train_state.TrainState) -> None:
    """Save model to disk."""
    os.makedirs(os.path.dirname(saved_model_path), exist_ok=True)
    if not saved_model_path.endswith(".msgpack"):
        saved_model_path += ".msgpack"
    bytes_output = flax.serialization.to_bytes(state)
    with open(saved_model_path, "wb") as f:
        f.write(bytes_output)
    print("Model saved to: ", saved_model_path)


def load_model(saved_model_path: str, state: train_state.TrainState
               ) -> train_state.TrainState:
    """Load model from disk."""
    with open(saved_model_path, "rb") as f:
        bytes_output = f.read()
    print("Model loaded from:", saved_model_path)
    return flax.serialization.from_bytes(state, bytes_output)


def create_train_state(module: nn.Module, rng: jax.Array, myconfig: munch.Munch
                       ) -> train_state.TrainState:
    """Creates an initial `TrainState`."""
    params = module.init(
        rng, jnp.ones([1, myconfig.model.seq_len, myconfig.model.n_mfcc])
    )["params"]
    tx = optax.adam(myconfig.train.learning_rate)
    return train_state.TrainState.create(
        apply_fn=module.apply, params=params, tx=tx)


def get_speaker_encoder(
    myconfig: munch.Munch,
    load_from: bool = None,
) -> tuple[BaseSpeakerEncoder, train_state.TrainState]:
    """Create speaker encoder model."""
    if myconfig.model.use_transformer:
        encoder = TransformerSpeakerEncoder(
            transformer_config=myconfig.model.transformer)
    else:
        encoder = LstmSpeakerEncoder(lstm_config=myconfig.model.lstm)

    tf.random.set_seed(0)
    init_rng = jax.random.PRNGKey(0)
    state = create_train_state(encoder, init_rng, myconfig)
    if load_from:
        state = load_model(load_from, state)

    return encoder, state


@partial(jax.jit, static_argnames=["batch_size", "triplet_alpha"])
def train_step(state: train_state.TrainState,
               batch_input: jax.Array,
               batch_size: int,
               triplet_alpha: float
               ) -> tuple[train_state.TrainState, jax.Array]:
    """Train for a single step."""
    def loss_fn(params: flax.core.frozen_dict.FrozenDict) -> jax.Array:
        # Compute loss.
        batch_output = state.apply_fn({'params': params}, batch_input)
        loss = get_triplet_loss_from_batch_output(
            batch_output, batch_size, triplet_alpha)
        return loss
    loss_grad_fn = jax.value_and_grad(loss_fn)
    loss_val, grads = loss_grad_fn(state.params)
    state = state.apply_gradients(grads=grads)
    return state, loss_val


def train_network(spk_to_utts: dataset.SpkToUtts,
                  myconfig: munch.Munch,
                  pool: Optional[multiprocessing.pool.Pool] = None
                  ) -> list[float]:
    start_time = time.time()
    losses = []
    _, state = get_speaker_encoder(myconfig)

    # Train
    for step in range(myconfig.train.num_steps):
        # Build batched input.
        batch_input = feature_extraction.get_batched_triplet_input(
            spk_to_utts, myconfig, pool)

        state, loss = train_step(
            state,
            batch_input,
            myconfig.train.batch_size,
            myconfig.train.triplet_alpha)
        losses.append(loss)

        print("step:", step, "/", myconfig.train.num_steps, "loss:", loss)

        if (myconfig.model.saved_model_path and
                (step + 1) % myconfig.train.save_model_frequency == 0):
            checkpoint = myconfig.model.saved_model_path
            if checkpoint.endswith(".msgpack"):
                checkpoint = checkpoint[:-8]
            checkpoint += ".ckpt-" + str(step + 1) + ".msgpack"
            save_model(checkpoint, state)

    training_time = time.time() - start_time
    print("Finished training in", training_time, "seconds")
    if myconfig.model.saved_model_path:
        save_model(myconfig.model.saved_model_path, state)
    return losses


def visualize_losses(losses: list[float]) -> None:
    plt.plot(losses)
    plt.xlabel("step")
    plt.ylabel("loss")
    plt.show()


def run_training(myconfig: munch.Munch) -> None:
    if myconfig.data.train_csv:
        spk_to_utts = dataset.get_csv_spk_to_utts(
            myconfig.data.train_csv)
        print("Training data:", myconfig.data.train_csv)
    else:
        spk_to_utts = dataset.get_librispeech_spk_to_utts(
            myconfig.data.train_librispeech_dir)
        print("Training data:", myconfig.data.train_librispeech_dir)
    with multiprocessing.Pool(myconfig.train.num_processes) as pool:
        losses = train_network(spk_to_utts,
                               myconfig,
                               pool)
    visualize_losses(losses)


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) == 0:
        config_file = "myconfig.yml"
    elif len(args) == 1:
        config_file = args[0]
    else:
        raise ValueError("Expecting a single argument: config file path")
    with open(config_file) as f:
        myconfig = munch.Munch.fromYAML(f.read())
    run_training(myconfig)



import jax
import jax.numpy as jnp
from flax.training import train_state
from multiprocessing.pool import ThreadPool
import time
from typing import Optional
import munch
import sys

from flaxspeaker import dataset
from flaxspeaker import feature_extraction
from flaxspeaker import neural_net


def run_inference(features: jax.Array,
                  state: train_state.TrainState,
                  myconfig: munch.Munch
                  ) -> jax.Array:
    """Get the embedding of an utterance using the encoder."""
    if myconfig.eval.full_sequence_inference:
        # Full sequence inference.
        batch_input = jnp.expand_dims(
            features, axis=0)
        batch_output = state.apply_fn({'params': state.params}, batch_input)
        return batch_output[0, :]
    else:
        # Sliding window inference.
        sliding_windows = feature_extraction.extract_sliding_windows(
            features, myconfig)
        if not sliding_windows:
            return None
        batch_input = jnp.stack(sliding_windows)
        batch_output = state.apply_fn({'params': state.params}, batch_input)

        # Aggregate the inference outputs from sliding windows.
        aggregated_output = jnp.mean(batch_output, axis=0, keepdims=False)
        return aggregated_output


class TripletScoreFetcher:
    """Class for computing triplet scores with multi-processing."""

    def __init__(self,
                 spk_to_utts: dataset.SpkToUtts,
                 state: train_state.TrainState,
                 myconfig: munch.Munch):
        self.spk_to_utts = spk_to_utts
        self.state = state
        self.config = myconfig

    def __call__(self, i: int) -> tuple[list[int], list[float]]:
        """Get the labels and scores from a triplet."""
        anchor, pos, neg = feature_extraction.get_triplet_features(
            self.spk_to_utts, self.config.model.n_mfcc)
        anchor_embedding = run_inference(
            anchor, self.state, self.config)
        pos_embedding = run_inference(
            pos, self.state, self.config)
        neg_embedding = run_inference(
            neg, self.state, self.config)
        if ((anchor_embedding is None) or
            (pos_embedding is None) or
                (neg_embedding is None)):
            # Some utterances might be smaller than a single sliding window.
            return ([], [])
        triplet_labels = [1, 0]
        triplet_scores = [
            neural_net.cosine_similarity(anchor_embedding, pos_embedding),
            neural_net.cosine_similarity(anchor_embedding, neg_embedding)]
        print("triplets evaluated:", i, "/", self.config.eval.num_triplets)
        return (triplet_labels, triplet_scores)


def compute_scores(
        state: train_state.TrainState,
        spk_to_utts: dataset.SpkToUtts,
        myconfig: munch.Munch
) -> tuple[list[int], list[float]]:
    """Compute cosine similarity scores from testing data."""
    labels = []
    scores = []
    fetcher = TripletScoreFetcher(spk_to_utts, state, myconfig)
    # CUDA does not support multi-processing, so using a ThreadPool.
    with ThreadPool(myconfig.train.num_processes) as pool:
        while myconfig.eval.num_triplets > len(labels) // 2:
            label_score_pairs = pool.map(fetcher, range(
                len(labels) // 2, myconfig.eval.num_triplets))
            for triplet_labels, triplet_scores in label_score_pairs:
                labels += triplet_labels
                scores += triplet_scores
    print("Evaluated", len(labels) // 2, "triplets in total")
    return (labels, scores)


def compute_eer(labels: list[int], scores: list[float],
                eval_threshold_step: float
                ) -> tuple[Optional[float], Optional[float]]:
    """Compute the Equal Error Rate (EER)."""
    if len(labels) != len(scores):
        raise ValueError("Length of labels and scored must match")
    eer_threshold = None
    eer = None
    min_delta = 1
    threshold = 0.0
    while threshold < 1.0:
        accept = [score >= threshold for score in scores]
        fa = [a and (1-l) for a, l in zip(accept, labels)]
        fr = [(1-a) and l for a, l in zip(accept, labels)]
        far = sum(fa) / (len(labels) - sum(labels))
        frr = sum(fr) / sum(labels)
        delta = abs(far - frr)
        if delta < min_delta:
            min_delta = delta
            eer = (far + frr) / 2
            eer_threshold = threshold
        threshold += eval_threshold_step

    return eer, eer_threshold


def run_eval(myconfig: munch.Munch) -> None:
    """Run evaluation of the saved model on test data."""
    start_time = time.time()
    if myconfig.data.test_csv:
        spk_to_utts = dataset.get_csv_spk_to_utts(
            myconfig.data.test_csv)
        print("Evaluation data:", myconfig.data.test_csv)
    else:
        spk_to_utts = dataset.get_librispeech_spk_to_utts(
            myconfig.data.test_librispeech_dir)
        print("Evaluation data:", myconfig.data.test_librispeech_dir)
    _, state = neural_net.get_speaker_encoder(
        myconfig,
        myconfig.model.saved_model_path)
    labels, scores = compute_scores(
        state, spk_to_utts, myconfig)
    eer, eer_threshold = compute_eer(
        labels, scores, myconfig.eval.threshold_step)
    eval_time = time.time() - start_time
    print("Finished evaluation in", eval_time, "seconds")
    print("eer_threshold =", eer_threshold, "eer =", eer)


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) == 0:
        config_file = "myconfig.yml"
    elif len(args) == 1:
        config_file = args[0]
    else:
        raise ValueError("Expecting a single argument: config file path")
    with open(config_file) as f:
        myconfig = munch.Munch.fromYAML(f.read())
    run_eval(myconfig)



import random
import os
import glob
import csv


SpkToUtts = dict[str, list[str]]


def get_librispeech_spk_to_utts(data_dir: str) -> SpkToUtts:
    """Get the dict from speaker to list of utterances for LibriSpeech."""
    flac_files = glob.glob(os.path.join(data_dir, "*", "*", "*.flac"))
    spk_to_utts = dict()
    for flac_file in flac_files:
        basename = os.path.basename(flac_file)
        split_name = basename.split("-")
        spk = split_name[0]
        if spk not in spk_to_utts:
            spk_to_utts[spk] = [flac_file]
        else:
            spk_to_utts[spk].append(flac_file)
    return spk_to_utts


def get_csv_spk_to_utts(csv_file: str) -> SpkToUtts:
    """Get the dict from speaker to list of utterances from CSV file."""
    spk_to_utts = dict()
    with open(csv_file) as f:
        reader = csv.reader(f)
        for row in reader:
            if len(row) != 2:
                continue
            spk = row[0].strip()
            utt = row[1].strip()
            if spk not in spk_to_utts:
                spk_to_utts[spk] = [utt]
            else:
                spk_to_utts[spk].append(utt)
    return spk_to_utts


def get_triplet(spk_to_utts: SpkToUtts) -> tuple[str, str, str]:
    """Get a triplet of anchor/pos/neg samples."""
    pos_spk, neg_spk = random.sample(list(spk_to_utts.keys()), 2)
    # Retry if too few positive utterances.
    while len(spk_to_utts[pos_spk]) < 2:
        pos_spk, neg_spk = random.sample(list(spk_to_utts.keys()), 2)
    anchor_utt, pos_utt = random.sample(spk_to_utts[pos_spk], 2)
    neg_utt = random.sample(spk_to_utts[neg_spk], 1)[0]
    return (anchor_utt, pos_utt, neg_utt)
